#!/usr/bin/env rspec -cfd

require_relative '../../spec_helper'

require 'cztop/reactor/signal_handling'


describe CZTop::Reactor::SignalHandling do

	let( :including_class ) do
		class_instance = Class.new do

			def initialize
				@signals = []
			end

			attr_reader :signals

			def handle_signal( signal_name )
				self.signals << signal_name
			end

		end
		class_instance.include( described_class )
		return class_instance
	end


	it "can add a signal handler setup to a reactor for a block" do
		obj = including_class.new
		reactor = CZTop::Reactor.new

		obj.with_signal_handler( reactor, :USR1 ) do
			obj.log.debug( "Killing with USR1" )
			Process.kill( :USR1, Process.pid )
			reactor.poll_once
		end

		expect( obj.signals ).to eq([ :USR1 ])
	end


	it "raises if the including class doesn't provide a #handle_signal method" do
		test_class = Class.new
		test_class.include( described_class )

		obj = test_class.new
		reactor = CZTop::Reactor.new

		expect {
			obj.with_signal_handler( reactor, :USR1 ) do
				obj.log.debug( "Killing with USR1" )
				Process.kill( :USR1, Process.pid )
				reactor.poll_once
			end
		}.to raise_error( NotImplementedError, /unhandled signal USR1/i )
	end


	it "doesn't try to tear down signal handling if it didn't get set up" do
		error_class = Class.new( StandardError )
		allow( CZTop::Socket::PAIR ).to receive( :new ).
			and_raise( error_class.new("something bad happened") )

		obj = including_class.new
		reactor = CZTop::Reactor.new

		begin
			obj.with_signal_handler( reactor, :USR1 ) do
				fail "we shouldn't get here"
			end
		rescue error_class => err
			expect( err.message ).to eq( "something bad happened" )
		end
	end

end

